@@ -1,9 +1,12 @@
+import io
 from unittest import TestCase
 from unittest.mock import patch, MagicMock
 from eaasy.extensions import buil_model, build_resource, build_dynamic_class
 from flask_restx import Api, fields, Namespace, Model
+from flask_restx.reqparse import RequestParser
 from eaasy.domain.database import BaseEntity
 from flask import Flask
+from openpyxl import load_workbook, Workbook
 
 class MockProperty:
     def __init__(self, name, type, nullable):
@@ -246,8 +249,20 @@ class TestBuilder(TestCase):
         mock_entity.get_all.return_value = []
         mock_entity.get_by_id.return_value = {}
         mock_entity.create.return_value = {}
+        mock_entity.update.return_value = {}
+        mock_entity.__name__ = "TestBaseEntity"
 
         return app, namespace, mock_entity
+
+    def build_workbook(self, rows: list[list[object]]):
+        workbook = Workbook()
+        worksheet = workbook.active
+        for row in rows:
+            worksheet.append(row) # type: ignore
+        stream = io.BytesIO()
+        workbook.save(stream)
+        stream.seek(0)
+        return stream
     
     def test_getting_all_returns_not_found_when_resource_is_not_defined(self, *_):
         # Arrange
@@ -261,6 +276,305 @@ class TestBuilder(TestCase):
         # Assert
         self.assertEqual(404, response.status_code)
 
+    def test_exporting_returns_excel_file_when_resource_is_defined(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        mock_entity.get_all.return_value = [
+            TestBaseEntity(id=1, name="Alpha", email="alpha@example.com"),
+            TestBaseEntity(id=2, name="Beta", email="beta@example.com"),
+        ]
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_export=True,
+            file_headers=['id', 'name', 'email'],
+        )
+
+        # Act
+        with app.test_client() as client:
+            response = client.get(f'/{ns_name}/file/')
+
+        # Assert
+        self.assertEqual(200, response.status_code)
+        self.assertEqual(
+            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
+            response.headers['Content-Type']
+        )
+        self.assertIn('attachment; filename=TestBaseEntity.xlsx', response.headers['Content-Disposition'])
+
+        workbook = load_workbook(io.BytesIO(response.data))
+        sheet = workbook.active
+
+        header_values = [cell.value for cell in sheet[1]] # type: ignore
+        first_row_values = [cell.value for cell in sheet[2]] # type: ignore
+        second_row_values = [cell.value for cell in sheet[3]] # type: ignore
+
+        self.assertEqual(['id', 'name', 'email'], header_values)
+        self.assertEqual([1, 'Alpha', 'alpha@example.com'], first_row_values)
+        self.assertEqual([2, 'Beta', 'beta@example.com'], second_row_values)
+
+    def test_importing_creates_new_rows_when_unique_not_found(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        mock_entity.get_all.return_value = []
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_import=True,
+            file_unique_fields=['email'],
+        )
+        workbook_stream = self.build_workbook([
+            ['email', 'name'],
+            ['alpha@example.com', 'Alpha'],
+            ['beta@example.com', 'Beta'],
+        ])
+
+        # Act
+        with app.test_client() as client:
+            response = client.post(
+                f'/{ns_name}/file/',
+                data={'file': (workbook_stream, 'users.xlsx')},
+                content_type='multipart/form-data'
+            )
+
+        # Assert
+        self.assertEqual(200, response.status_code)
+        self.assertEqual(2, mock_entity.create.call_count)
+        mock_entity.update.assert_not_called()
+        self.assertEqual({'created': 2, 'updated': 0}, response.json) # type: ignore
+
+    def test_importing_updates_existing_rows_based_on_unique_fields(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        mock_entity.get_all.return_value = [
+            TestBaseEntity(id=1, email='alpha@example.com', name='Alpha'),
+        ]
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_import=True,
+            file_unique_fields=['email'],
+        )
+        workbook_stream = self.build_workbook([
+            ['email', 'name'],
+            ['alpha@example.com', 'Alpha Updated'],
+            ['gamma@example.com', 'Gamma'],
+        ])
+
+        # Act
+        with app.test_client() as client:
+            response = client.post(
+                f'/{ns_name}/file/',
+                data={'file': (workbook_stream, 'users.xlsx')},
+                content_type='multipart/form-data'
+            )
+
+        # Assert
+        self.assertEqual(200, response.status_code)
+        mock_entity.update.assert_called_once_with(1, email='alpha@example.com', name='Alpha Updated')
+        mock_entity.create.assert_called_once_with(email='gamma@example.com', name='Gamma')
+        self.assertEqual({'created': 1, 'updated': 1}, response.json) # type: ignore
+
+    def test_importing_returns_bad_request_when_file_missing(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_import=True,
+        )
+
+        # Act
+        with app.test_client() as client:
+            response = client.post(
+                f'/{ns_name}/file/',
+                data={},
+                content_type='multipart/form-data'
+            )
+
+        # Assert
+        self.assertEqual(400, response.status_code)
+        self.assertEqual('Missing file upload', response.json['message']) # type: ignore
+
+    def test_importing_registers_file_upload_parser(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_import=True,
+        )
+
+        file_resource = [r for r in namespace.resources if r.urls[0] == '/file/'][0]
+        post_doc = file_resource.resource.post.__apidoc__
+
+        self.assertIn('expect', post_doc)
+        self.assertIsInstance(post_doc['expect'][0], RequestParser)
+        parser = post_doc['expect'][0]
+        self.assertEqual('file', parser.args[0].name)
+        self.assertEqual('files', parser.args[0].location)
+
+    def test_preview_returns_table_rows(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_preview=True,
+        )
+        workbook_stream = self.build_workbook([
+            ['email', 'name'],
+            ['alpha@example.com', 'Alpha'],
+            ['beta@example.com', 'Beta'],
+        ])
+
+        # Act
+        with app.test_client() as client:
+            response = client.post(
+                f'/{ns_name}/file/preview/',
+                data={'file': (workbook_stream, 'users.xlsx')},
+                content_type='multipart/form-data'
+            )
+
+        # Assert
+        self.assertEqual(200, response.status_code)
+        self.assertEqual({
+            'headers': ['email', 'name'],
+            'rows': [
+                {'email': 'alpha@example.com', 'name': 'Alpha'},
+                {'email': 'beta@example.com', 'name': 'Beta'},
+            ]
+        }, response.json) # type: ignore
+
+    def test_preview_returns_bad_request_when_file_missing(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_preview=True,
+        )
+
+        # Act
+        with app.test_client() as client:
+            response = client.post(
+                f'/{ns_name}/file/preview/',
+                data={},
+                content_type='multipart/form-data'
+            )
+
+        # Assert
+        self.assertEqual(400, response.status_code)
+        self.assertEqual('Missing file upload', response.json['message']) # type: ignore
+
+    def test_exporting_without_headers_infers_from_entities(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        mock_entity.get_all.return_value = [
+            {'id': 1, 'name': 'Alpha'},
+            {'id': 2, 'name': 'Beta', 'extra': 'value'},
+        ]
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_export=True,
+            file_name='custom.xlsx'
+        )
+
+        # Act
+        with app.test_client() as client:
+            response = client.get(f'/{ns_name}/file/')
+
+        # Assert
+        self.assertEqual(200, response.status_code)
+        self.assertIn('attachment; filename=custom.xlsx', response.headers['Content-Disposition'])
+
+        workbook = load_workbook(io.BytesIO(response.data))
+        sheet = workbook.active
+
+        header_values = [cell.value for cell in sheet[1]] # type: ignore
+        first_row_values = [cell.value for cell in sheet[2]] # type: ignore
+        second_row_values = [cell.value for cell in sheet[3]] # type: ignore
+
+        self.assertEqual(['id', 'name', 'extra'], header_values)
+        self.assertEqual([1, 'Alpha', None], first_row_values)
+        self.assertEqual([2, 'Beta', 'value'], second_row_values)
+
+    def test_exporting_with_no_results_returns_empty_sheet(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        mock_entity.get_all.return_value = []
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_export=True,
+        )
+
+        # Act
+        with app.test_client() as client:
+            response = client.get(f'/{ns_name}/file/')
+
+        # Assert
+        self.assertEqual(200, response.status_code)
+        workbook = load_workbook(io.BytesIO(response.data))
+        sheet = workbook.active
+
+        self.assertEqual(1, sheet.max_row) # type: ignore
+        self.assertEqual(1, sheet.max_column) # type: ignore
+        self.assertIsNone(sheet.cell(row=1, column=1).value) # type: ignore
+
+    def test_exporting_returns_error_when_entity_raises_exception(self, *_):
+        # Arrange
+        ns_name = 'testentity'
+        app, namespace, mock_entity = self.setup_namespace(ns_name)
+        mock_entity.get_all.side_effect = Exception({
+            'status_code': 500,
+            'message': 'Internal Error',
+            'data': {},
+        })
+        build_resource(
+            mock_entity,
+            namespace,
+            Model(ns_name),
+            Model(ns_name),
+            file_export=True,
+        )
+
+        # Act
+        with app.test_client() as client:
+            response = client.get(f'/{ns_name}/file/')
+
+        # Assert
+        self.assertEqual(500, response.status_code)
+        self.assertEqual('Internal Error', response.json['message']) # type: ignore
+
     def test_getting_all_returns_ok_when_resource_is_defined(self, *_):
         # Arrange
         ns_name = 'testentity'
